# This file is part of pymadcad,  distributed under license LGPL v3
'''
	The functions here generare meshes from simpler objects (like lower dimension meshes, or simple primitives). You won't find here iterative methods, nor adaptative geometry operations, nor mesh modification operations.
	The functions here may eventually be complex but will always feel predictable and simple from a human perspective.
'''

from .mathutils import *
from .primitives import Axis
from .mesh import Mesh, Web, Wire, edgekey, facekeyo, MeshError, web, wire, suites, mkquad, mktri
from . import triangulation
from . import settings
from . import primitives
from .nprint import nprint

from copy import copy


__all__ = [
	'extrans', 'extrusion', 'revolution', 'saddle', 'tube', 
	'repeat', 'thicken', 'inflate', 'inflate_offsets', 'expand',
	'flatsurface', 'icosurface', 'subdivide',
	'square', 'brick', 'cylinder', 'cone', 'pyramid', 'icosahedron', 'icosphere', 'uvsphere', 'regon', 
	]

	
# --- extrusion things ---


def extrusion(trans, line: Web, alignment:float=0) -> Mesh:
	''' create a surface by extruding the given outline by a transformation
		
		Parameters:
			line:         a line (Web or Wire) or a surface (Mesh) to extrude
			trans:        any transformation object accepted by `mathutils.transform`
			alignment:	   when > 0 the line is extruded both sides (the transformation is linearly interpoled)
	'''
	trans = transform(trans)
	neutral = mat4()
	return extrans(line, [
				neutral + (trans-neutral)*(-alignment),
				neutral + (trans-neutral)*(1-alignment),
				],
				[(0,1,0)])

def revolution(angle: float, axis: Axis, profile: Web, resolution=None) -> Mesh:
	''' create a revolution surface by extruding the given outline
		`steps` is the number of steps between the start and the end of the extrusion
		
		Parameters:
			angle:    angle of rotation between the given profile and the final produced profile
			axis:     the axis to rotate around
			profile:  the shape to extrude
			resolution:   resolution setting for the biggest produced circle, such as for primitives
	'''
	if not isinstance(profile, (Mesh,Web,Wire)):	
		profile = web(profile)
	# get the maximum radius, to compute the curve resolution
	radius = 0
	for pt in profile.points:
		v = pt-axis[0]
		v -= project(v,axis[1])
		radius = max(radius, length(v))
	steps = settings.curve_resolution(abs(angle*radius), abs(angle), resolution)+2
	# use extrans
	def trans():
		for i in range(steps):
			yield rotatearound(i/(steps-1)*angle, axis)
	def links():
		for i in range(steps-2):          yield (i,i+1, 0)
		if abs(angle-2*pi) <= NUMPREC:    yield (steps-2, 0, 0)
		else:                             yield (steps-2, steps-1, 0)
	return extrans(profile, trans(), links())

def saddle(web1: Web, web2: Web) -> Mesh:
	''' create a surface by extruding outine1 translating each instance to the next point of outline2
	'''
	web1, web2 = web(web1), web(web2)
	def trans():
		s = web2.points[0]
		for p in web2.points:
			yield translate(mat4(1), p-s)
	return extrans(web1, trans(), ((*e,t)  for e,t in zip(web2.edges, web2.tracks)))

def tube(outline: Web, path: Wire, end=True, section=True) -> Mesh:
	''' create a tube surface by extrusing the outline along the path
		if `section` is True, there is a correction of the segments to keep the section undeformed by the curve
	'''
	path = wire(path)
	def trans():
		lastrot = quat()
		l = len(path)-1
		yield mat4(1)
		for i in range(1,l+1):
			
			if i < l:
				# calculer l'angle a ce coude
				o = path[i]
				v1 = normalize(path[i-1]-o)
				v2 = normalize(path[i+1]-o)
				c = cross(-v1,v2)
				o = dot(-v1,v2)
				cl = length(c)
				cn = c/cl
				ha = atan2(cl,o)/2
				hrot = angleAxis(ha, cn)
				# calcul de la rotation de la section
				rot = hrot * lastrot
				lastrot = hrot * rot
				m = mat3_cast(rot)
				# deformation de la section pour ne pas réduire le volume
				if section:	
					m = scaledir(normalize(v1+v2), 1/cos(ha)) * m
			else:
				m = mat3_cast(lastrot)
			# construction de la transformation
			yield transform(path[i]) * mat4(m) * transform(-path[0])
	
	trans = trans()
	if not end:		next(trans)
	# handle wires with no tracks
	if path.tracks:		links = ((i, i+1, path.tracks[i]) for i in range(len(path)-1))
	else:				links = ((i, i+1, 0) for i in range(len(path)-1))
	# generate
	return extrans(outline, trans, links)


def extrans(section, transformations, links) -> Mesh:
	''' create a surface by extruding and transforming the given outline.
		
		Parameters:
			section:           a `Web` or a `Mesh`
			transformations:   iterable of mat4, one each section
			link:              iterable of tuples (a,b,t)  with:
									`(a,b)` the sections to link (indices of values returned by `transformation`).
									`t` the group number of that link, to combine with the section groups		
	'''
	# prepare
	if isinstance(section, Mesh):
		face = copy(section)
		face.strippoints()
		section = face.outlines()
	else:
		face = None
		section = copy(web(section))
	reindex = section.strippoints()
	
	mesh = Mesh()	  # result mesh
	extremities = {}  # index of extremities in links graph (similar to Web extremities), associated to a boolean telling the direction of that extremity
	kept = {}         # transformations kept in memory, indexed by their sections
	groups = {}       # groups created by the extransion
	
	l = len(section.points)
	
	# generate all sections faces using links
	for a,b,u in links:
		al = a*l
		bl = b*l
		for (c,d),v in zip(section.edges, section.tracks):
			mesh.faces.append((al+c, al+d, bl+d))
			mesh.faces.append((al+c, bl+d, bl+c))
			t = groups.setdefault((u,v), len(groups))
			mesh.tracks.append(t)
			mesh.tracks.append(t)
		# find extremities
		if face:
			if a in extremities:   del extremities[a]
			else:	               extremities[a] = True
			if b in extremities:   del extremities[b]
			else:                  extremities[b] = False
	
	# generate all combined groups
	mesh.groups = [None] * len(groups)
	for (u,v), t in groups.items():
		mesh.groups[t] = section.groups[v]	# NOTE will change in the future to mention both u and v
	
	# generate all sections points using transformations
	for k,trans in enumerate(transformations):
		for p in section.points:
			mesh.points.append(vec3(trans*vec4(p,1)))
		# keep extremities transformations
		if face and k in extremities:
			kept[k] = trans
	
	# append faces at extremities
	if face:
		merges = {}  # point merge dictionnary at faces insertion
		
		for k, orient in extremities.items():
			for src,dst in enumerate(reindex):
				if dst >= 0:
					merges[src + len(mesh.points)] = dst + l*k
			end = face .transform(kept[k])
			mesh += end if extremities[k] else end.flip()
		mesh.mergepoints(merges)
	
	mesh.check()
	return mesh
	
def linstep(start, stop, x):
	if x <= start:	return 0
	if x >= stop:	return 1
	return (x-start)/(stop-start)

def inflate_offsets(surface: Mesh, offset: float, method='face') -> '[vec3]':
	''' displacements vectors for points of a surfaceace we want to inflate.
		
		Parameters:
			offset:
				the distance from the surface to the offseted surface. its meaning depends on `method`
			method:     
				determines if the distance is from the old to the new faces, edges or points
				possible values: `'face', 'edge', 'point'`
	'''
	pnormals = surface.vertexnormals()
	
	# smooth normal offsets laterally when they are closer than `offset`
	outlines = surface.outlines_oriented()
	l = len(pnormals)
	normals = deepcopy(pnormals)
	for i in range(5):	# the number of steps is the diffusion distance through the mesh
		for a,b in outlines:
			d = surface.points[a] - surface.points[b]		# edge direction
			t = cross(pnormals[a]+pnormals[b], d)	# surface tangent normal to the edge
			# contribution stars when the offseted points are closer than `offset`
			contrib = 1 - smoothstep(0, offset, length(offset*(pnormals[a]-pnormals[b])+d))
			normals[a] += contrib * 0.5*project(pnormals[b]-pnormals[a], t)
			normals[b] += contrib * 0.5*project(pnormals[a]-pnormals[b], t)
		# renormalize
		for i in range(l):
			pnormals[i] = normals[i] = normalize(normals[i])
	
	# compute offset length depending on the method
	if method == 'face':
		lengths = [inf]*len(pnormals)
		for face in surface.faces:
			fnormal = surface.facenormal(face)
			for p in face:
				lengths[p] = min(lengths[p], 1/dot(pnormals[p], fnormal))
		return typedlist((pnormals[p]*lengths[p]*offset   for p in range(len(pnormals))), dtype=vec3)
	
	elif method == 'edge':
		lengths = [inf]*len(pnormals)
		for edge,enormal in surface.edgenormals().items():
			for p in edge:
				lengths[p] = min(lengths[p], 1/dot(pnormals[p], enormal))
		return typedlist((pnormals[p]*lengths[p]*offset	for p in range(len(pnormals))), dtype=vec3)
		
	elif method == 'point':
		return typedlist((pnormals[p]*offset	for p in range(len(pnormals))), dtype=vec3)

def inflate(surface:Mesh, offset:float, method='face') -> 'Mesh':
	''' move all points of the surface to make a new one at a certain distance of the last one

		Parameters:
			offset:       the distance from the surface to the offseted surface. its meaning depends on `method`
			method:       determines if the distance is from the old to the new faces, edges or points
	'''
	return Mesh(
				typedlist((p+d   for p,d in zip(surface.points, inflate_offsets(surface, offset, method))), dtype=vec3),
				surface.faces,
				surface.tracks,
				surface.groups)

def thicken(surface: Mesh, thickness: float, alignment:float=0, method='face') -> 'Mesh':
	''' thicken a surface by extruding it, points displacements are made along normal. 

		Parameters:
			thickness:    determines the distance between the two surfaces (can be negative to go the opposite direction to the normal).
			alignment:    specifies which side is the given surface: 0 is for the first, 1 for the second side, 0.5 thicken all apart the given surface.
			method:       determines if the thickness is from the old to the new faces, edges or points
	'''
	displts = inflate_offsets(surface, thickness, method)
	
	a = alignment
	b = alignment-1
	m = (	Mesh(
				typedlist((p+d*a  for p,d in zip(surface.points,displts)), dtype=vec3), 
				surface.faces[:], 
				surface.tracks[:], 
				surface.groups)
		+	Mesh(
				typedlist((p+d*b  for p,d in zip(surface.points,displts)), dtype=vec3), 
				surface.faces, 
				surface.tracks, 
				surface.groups[:]
				) .flip() 
		)
	t = len(m.groups)
	l = len(surface.points)
	m.groups.append(None)
	for e in surface.outlines_oriented():
		mkquad(m, (e[0], e[1], e[1]+l, e[0]+l), t)
	return m


def expand(surface: Mesh, offset: float, collapse=True) -> Mesh:
	''' generate a surface expanding the input mesh on the tangent of the ouline neighboring faces
	
		Parameters:
			offset:		distance from the outline point to the expanded outline points
			collapse:	if True, expanded points leading to crossing edges will collapse into one
	'''
	# outline with associated face normals
	pts = surface.points
	edges = {}
	for face in surface.faces:
		for e in ((face[0], face[1]), (face[1], face[2]), (face[2],face[0])):
			if e in edges:	del edges[e]
			else:			edges[(e[1], e[0])] = surface.facenormal(face)
	
	# return the point on tangent for a couple of edges from the frontier
	def tangent(e0, e1):
		mid = axis_midpoint(
				(pts[e0[1]], pts[e0[0]] - pts[e0[1]]), 
				(pts[e1[0]], pts[e1[1]] - pts[e1[0]]),
				)
		d0 = pts[e0[1]] - pts[e0[0]]
		d1 = pts[e1[1]] - pts[e1[0]]
		n0, n1 = edges[e0], edges[e1]
		t = normalize(cross(n0, n1) + NUMPREC*(d0*length(d1)-d1*length(d0)) + NUMPREC**2 * (cross(n0, d0) + cross(n1, d1)))
		if dot(t, cross(n0, d0)) < 0:
			t = -t
		return mid + t * offset
	
	# cross neighbooring normals
	for loop in suites(edges, cut=False):
		assert loop[-1] == loop[0],  "non-manifold input mesh"
		loop.pop()
		# compute the offsets, and remove anticipated overlapping geometries
		extended = [None]*len(loop)
		for i in range(len(loop)):
			# consecutive edges around i-1
			ei0 = (loop[i-2], loop[i-1])
			ei1 = (loop[i-1], loop[i])
			ti = tangent(ei0, ei1)
			if collapse:
				tk = deepcopy(ti)
				weight = 1
				# j is moving to find how much points to gather
				ej0 = ei0
				for j in reversed(range(i-len(loop)+1, i-1)):
					# consecutive edges aroung j
					ej0, ej1 = (loop[j-1], loop[j]), ej0
					tj = tangent(ej0, ej1)
					
					if dot(ti - tj, pts[ei1[0]] - pts[ej0[1]]) <= NUMPREC * length2(pts[ei1[0]] - pts[ej0[1]]):
						tk += tj
						weight += 1
					else:
						break
				# store tangents
				for k in range(j+1, i):
					extended[k] = tk/weight
			else:
				extended[i-1] = ti
		
		# insert the new points
		j = l = len(pts)
		g = len(surface.groups)
		surface.groups.append(None)
		for i in range(len(extended)):
			if extended[i] != extended[i-1]:
				pts.append(extended[i])
				
		# create the faces
		for i in range(len(extended)):
			if extended[i] != extended[i-1]:
				mkquad(surface, (loop[i-1], loop[i], j, (j if j > l else len(pts)) -1), g)
				j += 1
			else:
				mktri(surface, (loop[i-1], loop[i], (j if j > l else len(pts)) -1), g)
	
	return surface

def axis_midpoint(a0: Axis, a1: Axis, x=0.5) -> vec3:
	''' return the midpoint of two axis. 
		`x` is the blending factor between `a0` and `a1`
		
		- `x = 0` gives the point of `a0` the closest to `a1`
		- `x = 1` gives the point of `a1` the closest to `a0`
	'''
	p0, d0 = a0
	p1, d1 = a1
	if dot(d0,d1)**2 == length2(d0)*length2(d1):
		return mix(p0, p1, 0.5)
	return mix(
		p0 + unproject(project(p1-p0, noproject(d0, d1)), d0),
		p1 + unproject(project(p0-p1, noproject(d1, d0)), d1),
		x)


# --- filling things ---

def flatsurface(outline, normal=None) -> 'Mesh':
	''' generates a surface for a flat outline using the prefered triangulation method .
	
		if `normal` is specified, it must be the normal vector to the plane, and will be used to orient the face.
	'''
	if isinstance(outline, Wire):
		m = triangulation.triangulation_outline(outline, normal)
	else:
		m = triangulation.triangulation(web(outline), normal)
	if normal and dot(m.facenormal(0), normal) < 0:
		m = m.flip()
	return m


def icosurface(pts, ptangents, resolution=None) -> 'Mesh':
	''' generate a surface ICO (a subdivided triangle) with its points interpolated using interpol2tri.
	
		- If normals are given instead of point tangents (for ptangents), the surface will fit a sphere.
		- Else ptangents must be a list of couples (2 edge tangents each point).
	'''
	# compute normals to points
	if isinstance(ptangents[0], tuple):
		normals = [None]*3
		for i in range(3):
			normals[i] = normalize(cross(ptangents[i][0], ptangents[i][1]))
	else:
		# if normals are given instead of tangents, compute tangents to fit a sphere surface
		normals = ptangents
		ptangents = [None]*3
		for i in range(3):
			ptangents[i] = (
				normalize(noproject(pts[i-2]-pts[i], normals[i])) * arclength(pts[i], pts[i-2], normals[i], normals[i-2]),
				normalize(noproject(pts[i-1]-pts[i], normals[i])) * arclength(pts[i], pts[i-1], normals[i], normals[i-1]),
				)
	
	# evaluate resolution (using a bad approximation of the length for now)
	div = max(( settings.curve_resolution(
					distance(pts[i-1], pts[i-2]), 
					anglebt(normals[i-1], normals[i-2]), 
					resolution)
				for i in range(3) ))
	
	return dividedtriangle(lambda u,v: intri_sphere(pts, ptangents, u,v), div)
	
def dividedtriangle(placement, div=1) -> 'Mesh':
	''' generate a subdivided triangle with points placed according to the placement closure
		`placement(a,b) -> vec3`
		with a,b such as a+b+c = 1 with a,b,c within [0;1]
	'''
	segts = div+2
	# place points
	mesh = Mesh(groups=[None])
	for i in range(segts):
		u = i/(segts-1)				
		for j in range(segts-i):
			v = j/(segts-1)
			p = placement(u,v)
			mesh.points.append(p)
	# create faces
	c = 0
	for i in reversed(range(1,segts+1)):
		for j in range(i-1):
			s = c+j
			mesh.faces.append((s, s+i, s+1))
		for j in range(1,i-1):
			s = c+j
			mesh.faces.append((s, s+i-1, s+i))
		c += i
	mesh.tracks = typedlist.full(0, len(mesh.faces), 'I')

	return mesh


def subdivide(mesh, div=1) -> 'Mesh':
	''' subdivide all faces by the number of cuts '''
	n = div+2
	pts = typedlist(dtype=vec3)
	faces = typedlist(dtype=uvec3)
	tracks = typedlist(dtype='I')
	c = 0
	for f,t in enumerate(mesh.tracks):
		# place the points
		o,p0,p1 = mesh.facepoints(f)
		x = p0-o
		y = p1-o
		for i in range(n):
			u = i/(n-1)	
			for j in range(n-i):
				v = j/(n-1)
				p = o + u*x + v*y
				pts.append(p)
		# create the faces
		for i in reversed(range(1,n+1)):
			for j in range(i-1):
				s = c+j
				faces.append(uvec3(s, s+i, s+1))
			for j in range(1,i-1):
				s = c+j
				faces.append(uvec3(s, s+i-1, s+i))
			c += i
		tracks.extend([t] * (len(faces)-len(tracks)))
	
	new = Mesh(pts, faces, tracks)
	new.mergeclose()
	return new


# --- standard shapes ---
	
def brick(*args, **kwargs) -> 'Mesh':
	''' a simple brick with rectangular sides 
	
		constructors
		
			- brick(Box)
			- brick(min, max)
			- brick(center=vec3(0), width=vec3(-inf))
	'''
	if len(args) == 1 and not kwargs and isinstance(args[0], Box):		
		box = args[0]
	else:							
		box = Box(*args, **kwargs)
	mesh = Mesh(
		[
			vec3(1, 0, 0),
			vec3(1, 0, 1),
			vec3(0, 0, 1),
			vec3(0, 0, 0),
			vec3(1, 1, 0),
			vec3(1, 1, 1),
			vec3(0, 1, 1),
			vec3(0, 1, 0)],
		[
			uvec3(0, 1, 2),
			uvec3(0, 2, 3),
			uvec3(4, 7, 6),
			uvec3(4, 6, 5),
			uvec3(0, 4, 5),
			uvec3(0, 5, 1),
			uvec3(1, 5, 6),
			uvec3(1, 6, 2),
			uvec3(2, 6, 7),
			uvec3(2, 7, 3),
			uvec3(4, 0, 3),
			uvec3(4, 3, 7)],
		[	0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5	],
		[None] * 6,
		)
	for i in range(len(mesh.points)):
		mesh.points[i] = mesh.points[i]*box.width + box.min
	return mesh
	
def cylinder(bottom:vec3, top:vec3, radius:float, fill=True) -> 'Mesh':
	''' create a revolution cylinder, with the given radius 
	
		Parameters:
		
			bottom, top (vec3): the cylinder extremities centers
			fill (bool):        whether to put faces at both extremities
	'''
	direction = top-bottom
	base = wire(primitives.Circle((bottom,normalize(direction)), radius))
	if fill:	base = flatsurface(base).flip()
	return extrusion(direction, base)

def cone(summit:vec3, base:vec3, radius:float, fill=True) -> 'Mesh':
	''' create a revolution cone, with a base of the given radius 
	
		Parameters:
			
			summit (vec3):  The point at the top of the cone
			base (vec3):    the center point of the base
			fill (bool):    whether to put a face at the base
	'''
	base = wire(primitives.Circle((base, normalize(summit-base)), radius))
	if fill:	base = flatsurface(base)
	return pyramid(summit, base)
		
def pyramid(summit:vec3, base) -> 'Mesh':
	''' create a pyramid with the given summit point and the given base 
	
		Parameters:
			summit (vec3):   the top (summit) of the cone, not necessarity in the center of the shape
			base: (Mesh,Web,Wire):  the base shape
	'''
	if isinstance(base, Mesh):
		outline = base.outlines().flip()
		outline.stripgroups()
		result = Mesh(base.points, groups=outline.groups)
	else:
		outline = web(base)
		result = Mesh(points=outline.points, groups=outline.groups)
	
	p = len(result.points)
	result.points.append(summit)
	for edge, track in zip(outline.edges, outline.tracks):
		result.faces.append(uvec3(edge, p))
		result.tracks.append(track)
	
	if isinstance(base, Mesh):
		result += base.flip()

	return result

def square(axis:primitives.Axis, width:float) -> 'Mesh':
	''' return a simple square with the given normal axis and square width.
		Useful to quickly create a cutplane
	'''
	x,y,z = dirbase(axis[1])
	return Mesh(
		typedlist([axis[0]+0.6*width*p   for p in ((x+y), (y-x), (-y-x), (-y+x))]),
		typedlist([uvec3(0,1,2), uvec3(2,3,0)]),
		groups=[None],
		)

def icosahedron(center:vec3, radius:float) -> 'Mesh':
	''' a simple icosahedron (see https://en.wikipedia.org/wiki/Icosahedron) '''
	phi = (1+ sqrt(5)) /2	# golden ratio
	m = Mesh(
		typedlist([
			vec3(0, 1, phi),
			vec3(1, phi, 0),
			vec3(phi, 0, 1),
			vec3(0, -1, phi),
			vec3(-1, phi, 0),
			vec3(phi, 0, -1),
			vec3(0, 1, -phi),
			vec3(1, -phi, 0),
			vec3(-phi, 0, 1),
			vec3(0, -1, -phi),
			vec3(-1, -phi, 0),
			vec3(-phi, 0, -1),
		]),
		typedlist([
			uvec3(0,1,4), uvec3(0,2,1), uvec3(0,3,2), uvec3(0,8,3), uvec3(0,4,8),
			uvec3(2,5,1), uvec3(1,6,4), uvec3(4,11,8), uvec3(8,10,3), uvec3(3,7,2),
			uvec3(3,10,7), uvec3(2,7,5), uvec3(1,5,6), uvec3(4,6,11), uvec3(8,11,10),
			uvec3(7,9,5), uvec3(5,9,6), uvec3(6,9,11), uvec3(11,9,10), uvec3(10,9,7),
		]),
		)
	f = radius/length(m.points[0])
	for i,p in enumerate(m.points):
		m.points[i] = f*p + center
	return m

def icosphere(center:vec3, radius:float, resolution=None) -> 'Mesh':
	''' a simple icosphere with an arbitrary resolution (see https://en.wikipedia.org/wiki/Geodesic_polyhedron).
	
		Points are obtained from a subdivided icosahedron and reprojected on the desired radius.
	'''
	ico = icosahedron(center, radius)
	div = settings.curve_resolution(2/6*pi*radius, 2/6*pi, resolution)
	ico = subdivide(ico, div-1)
	for i,p in enumerate(ico.points):
		ico.points[i] = center + radius * normalize(p-center)
	return ico

def uvsphere(center:vec3, radius:float, alignment=vec3(0,0,1), resolution=None) -> 'Mesh':
	''' a simple uvsphere (simple sphere obtained with a revolution of an arc) '''
	x,y,z = dirbase(alignment)
	mesh = revolution(2*pi, 
			(center, z),
			web(primitives.ArcCentered(
				(center,x), 
				center+radius*z, 
				center-radius*z, 
				resolution=resolution)),
			resolution=resolution)
	mesh.mergeclose()
	return mesh

def regon(axis:primitives.Axis, radius, n, alignment=None) -> 'Wire':
	''' create a regular n-gon `Wire`, the same way we create a `Circle` '''
	return primitives.Circle(axis, radius, 
				resolution=('div',n), 
				alignment=alignment or vec3(1,0,0),
				).mesh() .segmented()

def repeat(pattern, n:int, transform):
	''' create a mesh duplicating n times the given pattern, each time applying the given transform.
		
		Parameters:
		
			pattern:   can either be a `Mesh`, `Web` or `Wire`   
						the return type will depend on the input type
			n:         the number of repetitions
			transform:     is the transformation between each duplicate
	'''
	current = pattern
	pool = type(pattern)(groups=pattern.groups)
	if n:	pool += current
	for i in range(1,n):
		current = current.transform(transform)
		pool += current
	return pool





from .container import *
from .web import Web
from .wire import Wire


class Mesh(NMesh):
	''' set of triangles, used to represent volumes or surfaces.
		As volumes are represented by their exterior surface, there is no difference between representation of volumes and faces, juste the way we interpret it.
		
		Attributes:
			points:     typedlist of vec3 for points
			faces:		typedlist of uvec3 for faces, the triplet is (a,b,c) such that  cross(b-a, c-a) is the normal oriented to the exterior.
			tracks:	    typedlist of integers giving the group each face belong to
			groups:     custom information for each group
			options:	custom informations for the entire mesh
	'''
	__slots__ = 'points', 'faces', 'tracks', 'groups', 'options'
	
	# BEGIN --- special methods ---
	
	def __init__(self, points=None, faces=None, tracks=None, groups=None, options=None):
		self.points = ensure_typedlist(points, vec3)
		self.faces = ensure_typedlist(faces, uvec3)
		self.tracks = ensure_typedlist(tracks if tracks is not None else typedlist.full(0, len(self.faces), 'I'), 'I')
		self.groups = groups if groups is not None else [None] * (max(self.tracks, default=-1)+1)
		self.options = options or {}
	
	def __add__(self, other):
		''' return a new mesh concatenating the faces and points of both meshes '''
		if isinstance(other, Mesh):
			r = Mesh(
				self.points if self.points is other.points else self.points[:], 
				self.faces[:], 
				self.tracks[:], 
				self.groups if self.groups is other.groups else self.groups[:],
				)
			r.__iadd__(other)
			return r
		else:
			return NotImplemented
			
	def __iadd__(self, other):
		''' append the faces and points of the other mesh '''
		if isinstance(other, Mesh):		
			if self.points is other.points:
				self.faces.extend(other.faces)
			else:
				lp = len(self.points)
				self.points.extend(other.points)
				self.faces.extend(f+lp  for f in other.faces)
			if self.groups is other.groups:
				self.tracks.extend(other.tracks)
			else:
				lt = len(self.groups)
				self.groups.extend(other.groups)
				self.tracks.extend(track+lt  for track in other.tracks)
			return self
		else:
			return NotImplemented
		
	# END BEGIN --- data management ---
	
	def strippoints(self) -> list:
		''' remove points that are used by no faces, return the reindex list.
			if used is provided, these points will be removed without usage verification
			
			return a table of the reindex made
		'''
		self.points, self.faces, reindex = striplist(self.points, self.faces)
		return reindex
	
	def mergepoints(self, merges) -> 'self':
		''' merge points with the merge dictionnary {src index: dst index}
			merged points are not removed from the buffer.
		'''
		j = 0
		for i,f in enumerate(self.faces):
			f = uvec3(
				merges.get(f[0], f[0]),
				merges.get(f[1], f[1]),
				merges.get(f[2], f[2]),
				)
			if not (f[0] == f[1] or f[1] == f[2] or f[2] == f[0]):
				self.faces[j] = f
				self.tracks[j] = self.tracks[i]
				j += 1
		del self.faces[j:]
		del self.tracks[j:]
		return self
					
					
	# END BEGIN ---- mesh checks -----
		
	def issurface(self):
		''' return True if the mesh is a well defined surface (an edge has 2 connected triangles at maximum, with consistent normals)
			such meshes are usually called 'manifold'
		''' 
		reached = set()
		for face in self.faces:
			for e in ((face[0], face[1]), (face[1], face[2]), (face[2],face[0])):
				if e in reached:	return False
				else:				reached.add(e)
		return True
	
	def isenvelope(self):
		''' return True if the surfaces are a closed envelope (the outline is empty)
		'''
		return len(self.outlines_oriented()) == 0
	
	def check(self):
		''' raise if the internal data is inconsistent '''
		if not (isinstance(self.points, typedlist) and self.points.dtype == vec3):	raise MeshError("points must be a typedlist(dtype=vec3)")
		if not (isinstance(self.faces, typedlist) and self.faces.dtype == uvec3): 	raise MeshError("faces must be a typedlist(dtype=uvec3)")
		if not (isinstance(self.tracks, typedlist) and self.tracks.dtype == 'I'): 	raise MeshError("tracks must be a typedlist(dtype='I')")
		l = len(self.points)
		for face in self.faces:
			for p in face:
				if p >= l:	raise MeshError("some point indices are greater than the number of points", face, l)
			if face[0] == face[1] or face[1] == face[2] or face[2] == face[0]:	raise MeshError("some faces use the same point multiple times", face)
		if len(self.faces) != len(self.tracks):	raise MeshError("tracks list doesn't match faces list length")
		if max(self.tracks, default=-1) >= len(self.groups): raise MeshError("some face group indices are greater than the number of groups", max(self.tracks, default=-1), len(self.groups))
	

	# END BEGIN --- selection methods ---
	
	def groupnear(self, point) -> int:
		''' return group id if the face the closest to the given point '''
		return self.tracks[self.face_near(point)]
		
	def facenear(self, point) -> int:
		''' return the index of the closest triangle to the given point '''
		return min(	range(len(self.faces)), 
					lambda i: distance_pt(point, self.facepoints(i)) )
	
	def group(self, quals) -> 'Self':
		''' extract a part of the mesh corresponding to the designated groups.
			
			Groups can be be given in either the following ways:
				- a set of group indices
					
					This can be useful to combine with other functions. However it can be difficult for a user script to keep track of which index correspond to which created group
					
				- an iterable of group qualifiers
				
					This is the best way to designate groups, and is meant to be used in combination with `self.qual()`.
					This mode selects every group having all the input qualifiers
					
			Example:
			
				>>> # create a mesh with only the given groups
				>>> mesh.group({1, 3, 8, 9})   
				<Mesh ...>
				
				>>> # create a mesh with all the groups having the following qualifiers
				>>> mesh.group(['extrusion', 'arc'])   
				<Mesh ...>
		'''
		faces = typedlist(dtype=uvec3)
		tracks = typedlist(dtype='I')
		for i in self.qualified_indices(quals):
			faces.append(self.faces[i])
			tracks.append(self.tracks[i])
		return Mesh(self.points, faces, tracks, self.groups)
		
	def replace(self, mesh, groups=None) -> 'self':
		''' replace the given groups by the given mesh.
			If groups is not specified, it will take the matching groups (with same index) in the current mesh
		'''
		if groups:
			groups = set(self.qualified_groups(groups))
		else:
			groups = set(mesh.tracks)
		j = 0
		for i,t in enumerate(self.tracks):
			if t not in groups:
				self.faces[j] = self.faces[i]
				self.tracks[j] = t
				j += 1
		del self.faces[j:]
		del self.tracks[j:]
		self += mesh
		return self
	
	
	# END BEGIN --- extraction methods ---
		
	def facepoints(self, f) -> tuple:
		''' shorthand to get the points of a face (index is an int or a triplet) '''
		if isinstance(f, Integral):
			f = self.faces[f]
		return self.points[f[0]], self.points[f[1]], self.points[f[2]]
	
	def facenormal(self, f) -> vec3:
		''' normal for a face '''
		if isinstance(f, Integral):	
			f = self.faces[f]
		p0 = self.points[f[0]]
		e1 = self.points[f[1]] - p0
		e2 = self.points[f[2]] - p0
		return normalize(cross(e1, e2))
	
	def facenormals(self) -> '[vec3]':
		''' list normals for each face '''
		return typedlist(map(self.facenormal, self.faces), vec3)
	
	def edgenormals(self) -> '{uvec2: vec3}':
		''' dict of normals for each UNORIENTED edge '''
		normals = {}
		for face in self.faces:
			normal = self.facenormal(face)
			for edge in ((face[0], face[1]), (face[1], face[2]), (face[2],face[0])):
				e = edgekey(*edge)
				normals[e] = normals.get(e,0) + normal
		for e,normal in normals.items():
			normals[e] = normalize(normal)
		return normals
	
		
	def vertexnormals(self) -> '[vec3]':
		''' list of normals for each point '''
		
		# collect the mesh border as edges and as points
		outline = self.outlines_oriented()
		border = set()
		for a,b in outline:
			border.add(a)
			border.add(b)
		
		# sum contributions to normals
		l = len(self.points)
		normals = typedlist.full(vec3(0), l)
		for face in self.faces:
			normal = self.facenormal(face)
			if not isfinite(normal):	continue
			for i in range(3):
				o = self.points[face[i]]
				# point on the surface
				if face[i] not in border:
					# triangle normals are weighted by their angle at the point
					contrib = anglebt(self.points[face[i-2]]-o, self.points[face[i-1]]-o)
					normals[face[i]] += contrib * normal
				# point on the outline
				elif (face[i], face[i-1]) in outline:
					# only the triangle creating the edge does determine its normal
					normals[face[i]] += normal
					normals[face[i-1]] += normal
		
		for i in range(l):
			normals[i] = normalize(normals[i])
		assert len(normals) == len(self.points)
		return normals
		
	def tangents(self) -> '{int: vec3}':
		''' tangents to outline points '''
		# outline with associated face normals
		edges = {}
		for face in self.faces:
			for e in ((face[0], face[1]), (face[1], face[2]), (face[2],face[0])):
				if e in edges:	del edges[e]
				else:			edges[(e[1], e[0])] = self.facenormal(face)
		
		# cross neighbooring normals
		tangents = {}
		for loop in suites(edges, cut=False):
			assert loop[-1] == loop[0],  "non-manifold mesh"
			loop.pop()
			for i in range(len(loop)):
				c = cross(	edges[(loop[i-2],loop[i-1])], 
							edges[(loop[i-1],loop[i])] )
				o = cross(  self.points[loop[i-2]] - self.points[loop[i]],
				            edges[(loop[i-2],loop[i-1])] + edges[(loop[i-1],loop[i])] )
				tangents[loop[i-1]] = normalize(mix(o, c, clamp(length2(c)/length2(o)/NUMPREC, 0, 1) ))
		return tangents
	
	def edges(self) -> set:
		''' set of UNORIENTED edges present in the mesh '''
		edges = set()
		for face in self.faces:
			edges.add(edgekey(face[0], face[1]))
			edges.add(edgekey(face[1], face[2]))
			edges.add(edgekey(face[2], face[0]))
		return edges
	
	def edges_oriented(self) -> set:
		''' iterator of ORIENTED edges, directly retreived of each face '''
		for face in self.faces:
			yield face[0], face[1]
			yield face[1], face[2]
			yield face[2], face[0]
	
	def outlines_oriented(self) -> set:
		''' return a set of the ORIENTED edges delimiting the surfaces of the mesh '''
		edges = set()
		for face in self.faces:
			for e in ((face[0], face[1]), (face[1], face[2]), (face[2],face[0])):
				if e in edges:	edges.remove(e)
				else:			edges.add((e[1], e[0]))
		return edges
	
	def outlines_unoriented(self) -> set:
		''' return a set of the UNORIENTED edges delimiting the surfaces of the mesh 
			this method is robust to face orientation aberations
		'''
		edges = set()
		for face in self.faces:
			for edge in ((face[0],face[1]),(face[1],face[2]),(face[2],face[0])):
				e = edgekey(*edge)
				if e in edges:	edges.remove(e)
				else:			edges.add(e)
		return edges
	
	def outlines(self) -> 'Web':
		''' return a Web of ORIENTED edges '''
		return Web(self.points, self.outlines_oriented())
		
	def groupoutlines(self) -> 'Web':
		''' return a Web of ORIENTED edges indexing groups.
			
			On a frontier between multiple groups, there is as many edges as groups, each associated to a group.
		'''
		edges = typedlist(dtype=uvec2)		# outline
		tracks = typedlist(dtype='I')		# groups for edges

		tmp = {}	# faces adjacent to edges
		for i,face in enumerate(self.faces):
			for e in ((face[1],face[0]),(face[2],face[1]),(face[0],face[2])):
				track = self.tracks[i]
				if e in tmp:
					if tmp[e] != track:
						edges.append(e)
						tracks.append(track)
					del tmp[e]
				else:
					tmp[(e[1],e[0])] = track
		edges.extend(tmp.keys())
		tracks.extend(tmp.values())
		return Web(self.points, edges, tracks, self.groups)
		
	def frontiers(self, *args) -> 'Web':
		''' return a Web of UNORIENTED edges that split the given groups appart.
		
			The arguments are groups indices or lists of group qualifiers (as set in `qualify()`). If there is one only argument it is considered as as list of arguments.
		
			- if no argument is given, then return the frontiers between every groups
			- to include the groups edges that are on the group border but not at the frontier with an other group, add `None` to the group set
			
			Example:
			
				>>> m = Mesh([...], [uvec3(0,1,2), uvec3(2,1,3)], [0, 1], [...])
				>>> m.frontiers(0,1).edges
				[uvec2(1,2)]
				
				>>> # equivalent to
				>>> m.frontiers({0,1}).edges
				[uvec2(1,2)]
				
				>>> m.frontiers(0,None)
				[uvec2(0,1), uvec2(0,2)]
		'''
		if args:
			if len(args) == 1 and hasattr(args[0], '__iter__'):
				args = args[0]
			groups = set()
			for arg in args:
				if arg is None:		groups.add(None)
				else:				groups.update(self.qualified_groups(arg))
		else:
			groups = None
		
		edges = typedlist(dtype=uvec2)
		tracks = typedlist(dtype='I')
		couples = OrderedDict()
		belong = {}
		for i,face in enumerate(self.faces):
			if groups is None or self.tracks[i] in groups:	
				for edge in ((face[0],face[1]),(face[1],face[2]),(face[2],face[0])):
					e = edgekey(*edge)
					if e in belong:
						if belong[e] != self.tracks[i]:
							g = edgekey(belong[e],self.tracks[i])
							edges.append(e)
							tracks.append(couples.setdefault(g, len(couples)))
						del belong[e]
					else:
						belong[e] = self.tracks[i]
		if groups and None in groups:
			edges.extend(belong.keys())
		return Web(self.points, edges, tracks, list(couples))
	
	def surface(self) -> float:
		''' total surface of triangles '''
		s = 0
		for f in self.faces:
			a,b,c = self.facepoints(f)
			s += length(cross(a-b, a,c))/2
		return s
		
	def volume(self) -> float:
		''' return the volume enclosed by the mesh if composed of envelopes (else it has no meaning) '''
		o = self.barycenter()
		s = vec3(0)
		for f in self.faces:
			s += glm.determinant(mat3(self.points[f[0]] - o, self.points[f[1]] - o, self.points[f[2]] - o))
		return s
	
	def barycenter(self) -> vec3:
		''' surface barycenter of the mesh '''
		if not self.faces:	return vec3(0)
		acc = vec3(0)
		tot = 0
		for f in self.faces:
			a,b,c = self.facepoints(f)
			weight = length(cross(b-a, c-a))
			tot += weight
			acc += weight*(a+b+c)
		return acc / (3*tot)
			
	def propagate(self, atface, atisland=None, find=None, conn=None):
		''' return the unconnected parts of the mesh as several meshes '''
		if not conn:	
			conn = connef(self.faces)
		
		reached = [False] * len(self.faces)	# faces reached
		stack = []
		# procedure for finding the new islands to propagate on
		if not find:
			start = [0]
			def find(stack, reached):
				for i in range(start[0],len(reached)):
					if not reached[i]:
						stack.append(i)
						break
				start[0] = i
		# propagation
		while True:
			# search start point
			find(stack, reached)
			# end when everything reached
			if not stack:	break
			# propagate
			while stack:
				i = stack.pop()
				if reached[i]:	continue	# make sure this face has not been stacked twice
				reached[i] = True
				atface(i, reached)
				f = self.faces[i]
				for i in range(3):
					e = f[i],f[i-1]
					if e in conn and not reached[conn[e]]:
						stack.append(conn[e])
			if atisland:
				atisland(reached)
	
	
	def splitgroups(self, edges=None):
		''' split the mesh groups into connectivity separated groups.
			the points shared by multiple groups will be duplicated
			if edges is provided, only the given edges at group frontier will be splitted
			
			return a list of tracks for points
		'''
		if edges is None:	edges = self.frontiers().edges
		# mark points on the frontier
		frontier = [False]*len(self.points)
		for a,b in edges:
			frontier[a] = True
			frontier[b] = True
		# duplicate points and reindex faces
		points = copy(self.points)
		idents = typedlist.full(0, len(self.points), 'I')		# track id corresponding to each point
		duplicated = {}		# new point index for couples (frontierpoint, group)
		def repl(pt, track):
			if frontier[pt]:
				key = (pt,track)
				if key in duplicated:	i = duplicated[key]
				else:
					i = duplicated[key] = len(points)
					points.append(points[pt])
					idents.append(track)
				return i
			else:
				idents[pt] = track
				return pt
		faces = typedlist((uvec3(repl(a,t), repl(b,t), repl(c,t))  
						for (a,b,c),t in zip(self.faces, self.tracks)), dtype=uvec3)
		
		self.points = points
		self.faces = faces
		return idents
		
	def split(self, edges) -> 'Self':
		''' split the mesh around the given edges. 
			The points in common with two or more designated edges will be dupliated once or more, and the face indices will be reassigned so that faces each side of the given edges will own a duplicate of that point each.
		'''
		# get connectivity and set of edges to manage
		conn = connef(self.faces)
		edges = set(edgekey(*edge)   for edge in edges)
		# collect the points to multiply
		ranks = Counter(p  for e in edges for p in e)
		separations = set(p  for p, count in ranks.items() if count > 1)
		newfaces = deepcopy(self.faces)
		# for each edge, reassign neighboring faces to the proper points
		for edge in edges:
			for pivot in edge:
				if edge in conn and pivot in newfaces[conn[edge]]:
					dupli = len(self.points)
					self.points.append(self.points[pivot])
					
					# change the point index in every neighbooring face
					front = edge
					while front in conn:
						fi = conn[front]
						f = arrangeface(self.faces[fi], pivot)
						fm = arrangeface(newfaces[fi], pivot)
						
						assert f[0] == pivot
						if fm[0] != pivot: 
							break
						
						newfaces[fi] = uvec3(dupli, fm[1], fm[2])
						if pivot == front[0]:	front = (pivot, f[2])
						elif pivot == front[1]:	front = (f[1], pivot)
						else:
							raise AssertionError('error in connectivity')
						if edgekey(*front) in edges:
							break
		
		self.faces = newfaces
		return self
	
	def islands(self, conn=None) -> '[Mesh]':
		''' return the unconnected parts of the mesh as several meshes '''
		islands = []
		faces = typedlist(dtype=uvec3)
		tracks = typedlist(dtype='I')
		def atface(i, reached):
			faces.append(self.faces[i])
			tracks.append(self.tracks[i])
		def atisland(reached):
			islands.append(Mesh(self.points, deepcopy(faces), deepcopy(tracks), self.groups))
			faces.clear()
			tracks.clear()
		self.propagate(atface, atisland, conn=conn)
		return islands
	
	def flip(self) -> 'Self':
		''' flip all faces, getting the normals opposite '''
		return Mesh(self.points, 
					typedlist((uvec3(f[0],f[2],f[1]) for f in self.faces), dtype=uvec3), 
					self.tracks, 
					self.groups)
	
	def orient(self, dir=None, conn=None) -> 'Self':
		''' flip the necessary faces to make the normals consistent, ensuring the continuity of the out side.
			
			Argument `dir` tries to make the result deterministic:
			
				* if given, the outermost point in this direction will be considered pointing outside
				* if not given, the farthest point to the barycenter will be considered pointing outside
				
				note that if the mesh contains multiple islands, that direction must make sense for each single island
		'''
		if dir:	
			metric = lambda p, n: (dot(p, dir), abs(dot(n, dir)))
			orient = lambda p, n: dot(n, dir)
		else:	
			center = self.barycenter()
			metric = lambda p, n: (length2(p-center), abs(dot(n, p-center)))
			orient = lambda p, n: dot(n, p-center)
		if not conn:	
			conn = Asso(  (edgekey(*e),i)
							for i,f in enumerate(self.faces)
							for e in ((f[0],f[1]), (f[1],f[2]), (f[2],f[0]))
							)
		
		faces = self.faces
		normals = self.facenormals()
		
		reached = [False] * len(faces)	# faces reached
		stack = []
		
		# propagation
		while True:
			# search start point
			best = (-inf,0)
			candidate = None
			for i,f in enumerate(faces):
				if not reached[i]:
					for p in f:
						score = metric(self.points[p], normals[i])
						if score > best:
							best, candidate = score, i
							if orient(self.points[p], normals[i]) < 0:
								faces[i] = (f[2],f[1],f[0])
			# end when everything reached
			if candidate is None:
				break
			else:
				stack.append(candidate)
			# process neighbooring
			while stack:
				i = stack.pop()
				if reached[i]:	continue	# make sure this face has not been stacked twice
				reached[i] = True
				
				f = faces[i]
				for i in range(3):
					e = f[i], f[i-1]
					for n in conn[edgekey(*e)]:
						if reached[n]:	continue
						nf = faces[n]
						# check for orientation continuity
						if arrangeface(nf,f[i-1])[1] == f[i]:
							faces[n] = (nf[2],nf[1],nf[0])
						# propagate
						stack.append(n)
		
		return self
	
	# END BEGIN ----- output methods ------
	
	def display(self, scene):
		from .. import displays
		
		m = copy(self)
		
		m.split(m.frontiers().edges)
		edges = m.outlines().edges
		
		# select edges above a threshold
		tosplit = []
		thresh = cos(settings.display['sharp_angle'])
		conn = connef(m.faces)
		for edge, f1 in conn.items():
			if edge[1] > edge[0]:	continue
			f2 = conn.get((edge[1],edge[0]))
			if f2 is None:	continue
			if m.tracks[f1] != m.tracks[f2] or dot(m.facenormal(f1), m.facenormal(f2)) <= thresh:
				tosplit.append(edge)
		
		m.split(tosplit)
		
		# get the group each point belong to
		idents = [0] * len(m.points)
		for face, track in zip(m.faces, m.tracks):
			for p in face:
				idents[p] = track
		
		normals = m.vertexnormals()
		
		if not m.points or not m.faces:	
			return displays.Display()
		
		return displays.SolidDisplay(scene, 
				typedlist_to_numpy(m.points, 'f4'), 
				typedlist_to_numpy(normals, 'f4'), 
				typedlist_to_numpy(m.faces, 'u4'),
				typedlist_to_numpy(edges, 'u4'),
				typedlist_to_numpy(idents, 'u4'),
				color = self.options.get('color'),
				)
	
	def __repr__(self):
		return '<Mesh with {} points at 0x{:x}, {} faces>'.format(len(self.points), id(self.points), len(self.faces))
	
	def __str__(self):
		return 'Mesh(\n  points={},\n  faces={},\n  tracks={},\n  groups={},\n  options={})'.format(
					reprarray(self.points, 'points'),
					reprarray(self.faces, 'faces'),
					reprarray(self.tracks, 'tracks'),
					reprarray(self.groups, 'groups'),
					repr(self.options))
		
	# END






def mktri(mesh, pts, track=0):
	''' append a triangle '''
	mesh.faces.append(pts)
	mesh.tracks.append(track)

def mkquad(mesh, pts, track=0):
	''' append a quad, choosing the best diagonal '''
	if (	distance2(mesh.points[pts[0]], mesh.points[pts[2]]) 
		<=	distance2(mesh.points[pts[1]], mesh.points[pts[3]]) ):
		mesh.faces.append((pts[:-1]))
		mesh.faces.append((pts[3], pts[0], pts[2]))
	else:
		mesh.faces.append((pts[0], pts[1], pts[3]))
		mesh.faces.append((pts[2], pts[3], pts[1]))
	mesh.tracks.append(track)
	mesh.tracks.append(track)



# This file is part of pymadcad,  distributed under license LGPL v3

# cython: language_level=3, cdivision=True

from libc.math cimport fabs, ceil, floor, sqrt, fmod, INFINITY
cimport cython
import glm

cdef:
	DEF NUMPREC = 1e-13
	
	double pmod(double l, double r):
		''' proceed to a python-like modulo on floats 
			(C module doesn't give a good reminder for negative values) 
		'''
		m = fmod(l,r)
		if m < 0:	m += r
		return m
	
	struct cvec3:
		double x
		double y
		double z
	
	double * varr(cvec3 *v):
		return <double*> v

	double dot(cvec3 a, cvec3 b):
		return a.x*b.x + a.y*b.y + a.z*b.z
	cvec3 cross(cvec3 a, cvec3 b):
		return cvec3(a.y*b.z - a.z*b.y,   a.z*b.x - a.x*b.z,   a.x*b.y - a.y*b.x)
	cvec3 vabs(cvec3 v):
		return cvec3(fabs(v.x), fabs(v.y), fabs(v.z))
	double amax(double *v, size_t l):
		cdef double m
		m = v[0]
		for i in range(1,l):
			if v[i] > m:	m = v[i]
		return m
	double amin(double *v, size_t l):
		cdef double m
		m = v[0]
		for i in range(1,l):
			if v[i] < m:	m = v[i]
		return m
	double aimax(double *v, size_t l):
		j = 0
		for i in range(1,l):
			if v[i] > v[j]:	j = i
		return j
	double aimin(double *v, size_t l):
		j = 0
		for i in range(1,l):
			if v[i] < v[j]:	j = i
		return j
	double vmax(cvec3 v):
		return amax(<double*>&v,3)
	double vmin(cvec3 v):
		return amin(<double*>&v,3)
	cvec3 vadd(cvec3 a, cvec3 b):
		return cvec3(a.x+b.x, a.y+b.y, a.z+b.z)
	cvec3 vsub(cvec3 a, cvec3 b):
		return cvec3(a.x-b.x, a.y-b.y, a.z-b.z)
	cvec3 vmul(cvec3 v, double r):
		return cvec3(r*v.x, r*v.y, r*v.z)
	double length(cvec3 v):
		return sqrt(dot(v,v))
	double norminf(cvec3 v):
		cdef cvec3 a = cvec3(abs(v.x), abs(v.y), abs(v.z))
		return amax(<double*>&a,3)

	cvec3 normalize(cvec3 v):
		return vmul(v, 1/length(v))
		
	cvec3 vaffine(cvec3 b, cvec3 a, double x):
		return cvec3(b.x + a.x*x,
					b.y + a.y*x,
					b.z + a.z*x)

cdef cvec3 glm2c(v):
	cdef cvec3 r
	cdef size_t i
	assert isinstance(v, (glm.dvec3, glm.fvec3, glm.ivec3))
	cdef double[:] a = v
	for i in range(3):	varr(&r)[i] = a[i]
	return r

cdef object c2glm(cvec3 v):
	return glm.dvec3(v.x, v.y, v.z)
	
cdef int dsign(double v):
	if v > 0:	return 1
	elif v < 0:	return -1
	else:		return 0
	

	
cdef long key(double f, double cell):
	''' hashing key for a float '''
	return <long> floor(f/cell)
	
def rasterize_segment(spaceo, double cell):
	''' return a list of hashing keys for an edge '''
	cdef cvec3 n, v, o
	cdef size_t order[3]
	cdef size_t reorder[3]
	cdef size_t i,j,k
	cdef double x,y,z, xmin,ymin, zmin,zmax
	cdef long pk[3]
	cdef cvec3 space[2]
	cdef double prec
	
	if cell <= 0:	raise ValueError('cell must be strictly positive')
	
	space = [glm2c(spaceo[0]), glm2c(spaceo[1])]
	rasterization = []
	prec = NUMPREC * max(norminf(space[0]), norminf(space[1]))
	
	# permutation of coordinates to get the direction the closest to Z
	n = vabs(vsub(space[1], space[0]))
	if vmax(n) < prec:	return rasterization
	if   n.y >= n.x and n.y >= n.z:		order,reorder = [2,0,1],[1,2,0]
	elif n.x >= n.y and n.x >= n.z:		order,reorder = [1,2,0],[2,0,1]
	else:								order,reorder = [0,1,2],[0,1,2]
	for i in range(2):
		temp = space[i]
		for j in range(3):
			varr(&space[i])[j] = varr(&temp)[order[j]]
	
	# prepare variables
	v = vsub(space[1], space[0])
	cell2 = cell/2
	assert v.z, order
	dy = v.y/v.z
	dx = v.x/v.z
	o = space[0]
	
	# z selection
	zmin,zmax = space[0].z, space[1].z
	if v.z < 0:		zmin,zmax = zmax,zmin
	zmin -= prec
	zmax += prec
	zmin -= pmod(zmin,cell)
	for i in range(max(1,<size_t>ceil((zmax-zmin)/cell))):
		z = zmin + cell*i + cell2
		
		# y selection
		ymin,ymax = o.y + dx*(z-cell2-o.z),  o.y + dx*(z+cell2-o.z)
		if dy < 0:	ymin,ymax = ymax,ymin
		ymin -= prec
		ymax += prec
		ymin -= pmod(ymin,cell)
		for j in range(max(1,<size_t>ceil((ymax-ymin)/cell))):
			y = ymin + j*cell + cell2
			
			# x selection
			xmin,xmax = o.x + dx*(z-cell2-o.z),  o.x + dx*(z+cell2-o.z)
			if dx < 0:	xmin,xmax = xmax,xmin	
			xmin -= prec
			xmax += prec
			xmin -= pmod(xmin,cell)
			for k in range(max(1,<size_t>ceil((xmax-xmin)/cell))):
				x = xmin + k*cell + cell2
				
				pk = [key(x,cell), key(y,cell), key(z,cell)]
				rasterization.append(( pk[reorder[0]], pk[reorder[1]], pk[reorder[2]] ))
	return rasterization


def rasterize_triangle(spaceo, double cell):
	''' return a list of hashing keys for a triangle '''
	cdef size_t i,j,k
	cdef size_t order[3]
	cdef size_t reorder[3]
	cdef cvec3 v[3]
	cdef double candz[4]
	cdef double candy[6]
	cdef size_t candylen
	cdef long pk[3]
	cdef cvec3 pmin, pmax
	cdef double xmin,xmax, ymin,ymax, zmin,zmax
	cdef cvec3 space[3]
	cdef double prec
	
	if cell <= 0:	raise ValueError('cell must be strictly positive')
	
	space = [glm2c(spaceo[0]), glm2c(spaceo[1]), glm2c(spaceo[2])]
	rasterization = []
	prec = NUMPREC*max(norminf(space[0]), norminf(space[1]), norminf(space[2]))
	
	# permutation of coordinates to get the normal the closer to Z
	n = vabs(cross(vsub(space[1],space[0]), vsub(space[2],space[0])))
	if vmax(n) < prec:	return rasterization
	if   n.y >= n.x and n.y >= n.z:		order,reorder = [2,0,1],[1,2,0]
	elif n.x >= n.y and n.x >= n.z:		order,reorder = [1,2,0],[2,0,1]
	else:								order,reorder = [0,1,2],[0,1,2]
	for i in range(3):
		temp = space[i]
		for j in range(3):
			varr(&space[i])[j] = varr(&temp)[order[j]]
	
	# prepare variables
	# WARNING: due to C differences with modulo (%) we can't use the negative indices for arrays
	v = [vsub(space[0],space[1]), vsub(space[1],space[2]), vsub(space[2],space[0])]
	n = cross(v[0],v[1])
	assert n.z
	dx = -n.x/n.z
	dy = -n.y/n.z
	o = space[0]
	cell2 = cell/2
	pmin = cvec3(
			min(space[0].x, space[1].x, space[2].x),
			min(space[0].y, space[1].y, space[2].y),
			min(space[0].z, space[1].z, space[2].z),
			)
	pmax = cvec3(
			max(space[0].x, space[1].x, space[2].x),
			max(space[0].y, space[1].y, space[2].y),
			max(space[0].z, space[1].z, space[2].z),
			)
	xmin,xmax = pmin.x,pmax.x
	for i in range(3):	varr(&pmin)[i] -= pmod(varr(&pmin)[i], cell)
	for i in range(3):	varr(&pmax)[i] += cell - pmod(varr(&pmax)[i], cell)
	
	# x selection
	xmin -= prec
	xmax += prec
	xmin -= pmod(xmin,cell)
	for i in range(max(1,<size_t>ceil((xmax-xmin)/cell))):
		x = xmin + cell*i + cell2
	
		# y selection
		candylen = 0
		for i in range(3):
			# NOTE: cet interval ajoute parfois des cases inutiles apres les sommets
			if (space[(i+1)%3].x-x+cell2)*(space[i].x-x-cell2) <= 0 or (space[(i+1)%3].x-x-cell2)*(space[i].x-x+cell2) <= 0:
				d = v[i].y / (v[i].x if v[i].x else INFINITY)
				candy[candylen]   = ( space[i].y + d * (x-cell2-space[i].x) )
				candy[candylen+1] = ( space[i].y + d * (x+cell2-space[i].x) )
				candylen += 2
		ymin,ymax = max(pmin.y,amin(candy,candylen)), min(pmax.y,amax(candy,candylen))
		ymin -= prec
		ymax += prec
		ymin -= pmod(ymin,cell)
		if ymax < ymin:	continue
		for j in range(max(1,<size_t>ceil((ymax-ymin)/cell))):
			y = ymin + cell*j + cell2
		
			# z selection
			candz = [
				o.z + dx*(x-cell2-o.x) + dy*(y-cell2-o.y),
				o.z + dx*(x+cell2-o.x) + dy*(y-cell2-o.y),
				o.z + dx*(x-cell2-o.x) + dy*(y+cell2-o.y),
				o.z + dx*(x+cell2-o.x) + dy*(y+cell2-o.y),
				]
			zmin,zmax = max(pmin.z,amin(candz,4)), min(pmax.z,amax(candz,4))
			zmin -= prec
			zmax += prec
			zmin -= pmod(zmin,cell)
			if zmax < zmin:	continue
			for k in range(max(1,<size_t>ceil((zmax-zmin)/cell))):
				z = zmin + cell*k + cell2
				
				# remove box from corners that goes out of the area
				if pmin.x<x and pmin.y<y and pmin.z<z and x<pmax.x and y<pmax.y and z<pmax.z:
					pk = [key(x,cell), key(y,cell), key(z,cell)]
					rasterization.append(( pk[reorder[0]], pk[reorder[1]], pk[reorder[2]] ))
	return rasterization


def intersect_triangles(f0, f1, precision):
	''' Intersects 2 triangles and outputs intersections vertices

		f0 = first face (tuple of 3 vertices given in clock wise orientation. vertices are glm.vec3 or glm.dvec3)
		f1 = second face

		output = None if no intersection
					2 intersection vertices given as : 
					((fi, ej, xj), (fk, em, xm)) where
						fi, fj = face id
						ej, em = edge id on face
						xj, xm = intersection point (same precision as input vertices) 
						
		restrictions : vertices on faces must be spatially different. identical vertices on triangles are not managed
	'''
	cdef cvec3 A1A2, A1A3, B1B2, B1B3, nA, nB, d, xA, xB, yA, yB
	cdef int i

	cdef cvec3[3] fA = [glm2c(f0[0]), glm2c(f0[1]), glm2c(f0[2])]
	cdef cvec3[3] fB = [glm2c(f1[0]), glm2c(f1[1]), glm2c(f1[2])]
	cdef double prec = precision
	
	# get the normal to the first face
	A1A2 = vsub(fA[1],fA[0])
	A1A3 = vsub(fA[2],fA[0])
	nA = normalize(cross(A1A2, A1A3))
	
	# get the normal to the second face
	B1B2 = vsub(fB[1],fB[0])
	B1B3 = vsub(fB[2],fB[0])
	nB = normalize(cross(B1B2, B1B3))
	
	# gets the direction of the intersection between the plan containing fA and the one containing fB 
	d1 = cross(nA, nB)
	ld1 = length(d1)
	if ld1 <= prec :
		#print("coplanar or parallel faces")
		return None
	d = vmul(d1, 1/ld1)
	
	# projection direction on to d from fA and fB
	tA = cross(nA, d)
	tB = cross(nB, d)
	
	# project fA summits onto d (in pfA)
	# xA being the coordinates of fA onto d
	pA1 = vsub(fA[0],  vmul(tA, dot(vsub(fA[0],fB[0]), nB) / dot(tA,nB)) )
	xA = cvec3(0, dot(A1A2,d), dot(A1A3,d))
	cdef cvec3[3] pfA = [pA1, vaffine(pA1, d, xA.y), vaffine(pA1, d, xA.z)]
	
	# project fB summits onto d
	xB = cvec3(dot(vsub(fB[0],fA[0]), d), dot(vsub(fB[1],fA[0]), d), dot(vsub(fB[2],fA[0]), d))
	cdef cvec3[3] pfB = [vaffine(pA1, d, xB.x), vaffine(pA1, d, xB.y), vaffine(pA1, d, xB.z)]
	
	# project fA and fB summits on transversal direction tA and tB
	for i in range(3):
		varr(&yA)[i] = dot(vsub(fA[i], pfA[i]), tA)
		varr(&yB)[i] = dot(vsub(fB[i], pfB[i]), tB)
	
	# identify signs of yA and yB
	cdef int[3] sYA
	cdef int[3] sYB
	for i in range(3):
		if abs(varr(&yA)[i]) <= prec:
			sYA[i] = 0
			varr(&yA)[i] = 0
		else:
			sYA[i] = dsign(varr(&yA)[i])
		if abs(varr(&yB)[i]) <= prec:
			sYB[i] = 0
			varr(&yB)[i] = 0
		else:
			sYB[i] = dsign(varr(&yB)[i])
	
	# check if triangles have no intersections with line D
	if abs(sYA[0]+sYA[1]+sYA[2]) == 3 or abs(sYB[0]+sYB[1]+sYB[2]) == 3:
		#print("plans intersects but no edges intersection (1)")
		return None

	# we know that triangles do intersect the line D
	# edges of intersection on A and B with the convention : edge i of face X connects fX[i] and fX[(i+1)%3] 
	cdef int eIA[3]
	cdef int eIB[3]
	cdef size_t neIA=0
	cdef size_t neIB=0
	
	cdef int j, k
	# prioritize on edges really getting through the face (not stopping on)
	for j in range(3):
		if sYA[j]*sYA[(j+1)%3] < 0:
			break		
	# look for edges intersecting starting from the eventual through one
	for i in range(j,j+3):
		if sYA[i%3]*sYA[(i+1)%3] <= 0 and abs(sYA[i%3])+abs(sYA[(i+1)%3]) > 0 : 
			eIA[neIA] = i%3
			neIA += 1
		if sYB[i%3]*sYB[(i+1)%3] <= 0 and abs(sYB[i%3])+abs(sYB[(i+1)%3]) > 0 : 
			eIB[neIB] = i%3
			neIB += 1
	if neIA==1:		eIA[1] = eIA[0]
	if neIB==1:		eIB[1] = eIB[0]

	# intersections coordinates onto line D
	cdef double xIA[2]
	cdef double xIB[2]
	for i in range(2):
		xIA[i] = (varr(&yA)[(eIA[i]+1)%3] * varr(&xA)[eIA[i]] - varr(&yA)[eIA[i]] * varr(&xA)[(eIA[i]+1)%3]) / (varr(&yA)[(eIA[i]+1)%3] - varr(&yA)[eIA[i]])
		xIB[i] = (varr(&yB)[(eIB[i]+1)%3] * varr(&xB)[eIB[i]] - varr(&yB)[eIB[i]] * varr(&xB)[(eIB[i]+1)%3]) / (varr(&yB)[(eIB[i]+1)%3] - varr(&yB)[eIB[i]])
		
	# intervals of intersections
	piA, miA = (0, 1)	if xIA[0] > xIA[1] else   (1, 0)
	piB, miB = (0, 1)	if xIB[0] > xIB[1] else   (1, 0)
	
    # one intersection at the border of the intervals
	if abs(xIA[piA]-xIB[miB]) <= prec:
		# edge of max from A matches min of B
		return (0, eIA[piA], c2glm(vaffine(pA1, d, xIA[piA]))),  (1, eIB[miB], c2glm(vaffine(pA1, d, xIB[miB])))
		
	if abs(xIB[piB]-xIA[miA]) <= prec:
		# edge of max from B matches min of A
		return (0, eIA[miA], c2glm(vaffine(pA1, d, xIA[miA]))),  (1, eIB[piB], c2glm(vaffine(pA1, d, xIB[piB])))
	
	# no intersection - intervals doesn't cross
	if xIB[piB]-prec < xIA[miA] or xIA[piA]-prec < xIB[miB]:
		#print("plans intersects but no edges intersection (2)")
		return None
		
	# one interval is included in the other one
	if xIB[miB]-prec <= xIA[miA] and xIA[piA]-prec <= xIB[piB]:
		# edges of A cross face B
		return (0, eIA[miA], c2glm(vaffine(pA1, d, xIA[miA]))),  (0, eIA[piA], c2glm(vaffine(pA1, d, xIA[piA])))
	if xIA[miA]-prec <= xIB[miB] and xIB[piB]-prec <= xIA[piA]:
		# edges of A cross face B
		#return (1, eIB[miB], c2glm(vaffine(pA1, d, xIB[miB]))),  (1, eIB[piB], c2glm(vaffine(pA1, d, xIB[piB])))
		
		# give priority to face index 0 when equivalent regarding the precision
		if abs(xIA[miA]-xIB[miB]) <= prec:	mr = (0, eIA[miA], c2glm(vaffine(pA1, d, xIA[miA])))
		else:								mr = (1, eIB[miB], c2glm(vaffine(pA1, d, xIB[miB])))
		if abs(xIB[piB]-xIA[piA]) <= prec:	pr = (0, eIA[piA], c2glm(vaffine(pA1, d, xIA[piA])))
		else:								pr = (1, eIB[piB], c2glm(vaffine(pA1, d, xIB[piB])))
		return mr, pr
	
	# intervals cross each other
	if xIB[miB] > xIA[miA]-prec and xIA[piA]-prec < xIB[piB]:
		# M edge of B crosses face A and P edge of A crosses face B
		return (0, eIA[piA], c2glm(vaffine(pA1, d, xIA[piA]))), (1, eIB[miB], c2glm(vaffine(pA1, d, xIB[miB])))
	if xIA[miA] > xIB[miB]-prec and xIB[piB]-prec < xIA[piA]:
		# M edge of A crosses face B and P edge of B crosses face A
		return (0, eIA[miA], c2glm(vaffine(pA1, d, xIA[miA]))), (1, eIB[piB], c2glm(vaffine(pA1, d, xIB[piB])))
	
	raise Exception("unexpected case: {} {}".format(repr(fA), repr(fB)))


